from __future__ import annotations

import argparse
import sys
from pathlib import Path
from typing import Optional, Union

from dolphin._types import TropoModel, TropoType
from dolphin.utils import get_cpu_count

from .config import (
    DisplacementWorkflow,
    ShpMethod,
    UnwrapMethod,
)


def create_config(
    *,
    outfile: Union[str, Path],
    print_empty: bool = False,
    slc_files: Optional[list[str]] = None,
    subdataset: Optional[str] = None,
    keep_paths_relative: bool = False,
    work_directory: Optional[Path] = Path(),
    mask_file: Optional[str] = None,
    zero_where_masked: bool = False,
    ministack_size: Optional[int] = 15,
    half_window_size: tuple[int, int] = (11, 5),
    shp_method: ShpMethod = ShpMethod.GLRT,
    baseline_lag: Optional[int] = None,
    amp_dispersion_threshold: float = 0.25,
    strides: tuple[int, int],
    output_bounds: tuple[int, int, int, int],
    block_shape: tuple[int, int] = (512, 512),
    threads_per_worker: int = 4,
    n_parallel_bursts: int = 1,
    enable_gpu: bool = False,
    ntiles: tuple[int, int] = (1, 1),
    downsample_factor: tuple[int, int] = (1, 1),
    no_unwrap: bool = False,
    no_inversion: bool = False,
    n_parallel_unwrap: int = 1,
    run_goldstein: bool = False,
    run_interpolation: bool = False,
    unwrap_method: UnwrapMethod = UnwrapMethod.SNAPHU,
    troposphere_files: Optional[list[str]] = None,
    tropo_date_fmt: str = "%Y%m%d",
    tropo_package: str = "pyaps",
    tropo_model: TropoModel = TropoModel.ERA5,
    tropo_delay_type: TropoType = TropoType.COMB,
    ionosphere_files: Optional[list[str]] = None,
    geometry_files: Optional[list[str]] = None,
    dem_file: Optional[str] = None,
    single_update: bool = False,
    log_file: Optional[Path] = None,
    amplitude_mean_files: Optional[list[str]] = None,
    amplitude_dispersion_files: Optional[list[str]] = None,
):
    """Create a config for a displacement workflow."""
    output = sys.stdout if outfile == "-" else outfile
    if print_empty:
        DisplacementWorkflow.print_yaml_schema(output)
        return

    if amplitude_dispersion_files is None:
        amplitude_dispersion_files = []
    if amplitude_mean_files is None:
        amplitude_mean_files = []
    if geometry_files is None:
        geometry_files = []
    if ionosphere_files is None:
        ionosphere_files = []
    if troposphere_files is None:
        troposphere_files = []
    if single_update:
        # create only one interferogram from the first and last SLC images
        interferogram_network = {"indexes": [(0, -1)]}
        # Override the ministack size so that only one phase linking is run
        ministack_size = 1000
    else:
        interferogram_network = {}  # Use default

    cfg = DisplacementWorkflow(
        cslc_file_list=slc_files,
        work_directory=work_directory,
        keep_paths_relative=keep_paths_relative,
        mask_file=mask_file,
        input_options={
            "subdataset": subdataset,
        },
        interferogram_network=interferogram_network,
        output_options={
            "strides": {"x": strides[0], "y": strides[1]},
            "bounds": output_bounds,
        },
        phase_linking={
            "ministack_size": ministack_size,
            "half_window": {"x": half_window_size[0], "y": half_window_size[1]},
            "shp_method": shp_method,
            "baseline_lag": baseline_lag,
        },
        ps_options={
            "amp_dispersion_threshold": amp_dispersion_threshold,
        },
        unwrap_options={
            "unwrap_method": unwrap_method,
            "run_unwrap": not no_unwrap,
            "zero_where_masked": zero_where_masked,
            "run_goldstein": run_goldstein,
            "run_interpolation": run_interpolation,
            "n_parallel_jobs": n_parallel_unwrap,
            "snaphu_options": {
                "ntiles": ntiles,
            },
            "tophu_options": {
                "ntiles": ntiles,
                "downsample_factor": downsample_factor,
            },
        },
        timeseries_options={
            "run_inversion": not no_inversion,
        },
        correction_options={
            "troposphere_files": troposphere_files,
            "tropo_date_fmt": tropo_date_fmt,
            "tropo_package": tropo_package,
            "tropo_model": tropo_model,
            "tropo_delay_type": tropo_delay_type,
            "ionosphere_files": ionosphere_files,
            "geometry_files": geometry_files,
            "dem_file": dem_file,
        },
        worker_settings={
            "block_shape": block_shape,
            "n_parallel_bursts": n_parallel_bursts,
            "threads_per_worker": threads_per_worker,
            "gpu_enabled": enable_gpu,
        },
        log_file=log_file,
        amplitude_mean_files=amplitude_mean_files,
        amplitude_dispersion_files=amplitude_dispersion_files,
    )

    print(f"Saving configuration to {output!s}", file=sys.stderr)
    cfg.to_yaml(output)


def get_parser(subparser=None, subcommand_name="run"):
    """Set up the command line interface."""
    metadata = {
        "description": "Create a configuration file for a displacement workflow.",
        "formatter_class": argparse.ArgumentDefaultsHelpFormatter,
        # https://docs.python.org/3/library/argparse.html#fromfile-prefix-chars
        "fromfile_prefix_chars": "@",
    }
    if subparser:
        # Used by the subparser to make a nested command line interface
        parser = subparser.add_parser(subcommand_name, **metadata)
    else:
        parser = argparse.ArgumentParser(**metadata)

    # parser._action_groups.pop()
    parser.add_argument(
        "--print-empty",
        action="store_true",
        help="Flag to print a YAML file with only default filled to `outfile`.",
    )
    parser.add_argument(
        "-o",
        "--outfile",
        default="dolphin_config.yaml",
        help="Name of YAML configuration file to save to. Use '-' to write to stdout.",
    )
    # Get Inputs from the command line
    inputs = parser.add_argument_group("Input options")
    inputs.add_argument(
        "--slc-files",
        nargs=argparse.ZERO_OR_MORE,
        help="List the paths of all SLC files to include.",
    )
    inputs.add_argument(
        "--keep-paths-relative",
        action="store_true",
        help=(
            "Don't resolve file paths that are given as relative (useful for running"
            " inside docker). "
        ),
    )
    # Get the subdataset of the SLCs to use, if passing HDF5/NetCDF files
    inputs.add_argument(
        "-sds",
        "--subdataset",
        help="Subdataset to use from HDF5/NetCDF files.",
    )
    inputs.add_argument(
        "--amplitude-mean-files",
        nargs=argparse.ZERO_OR_MORE,
        help="Optional: List the paths of existing amplitude mean files.",
        default=[],
    )
    inputs.add_argument(
        "--amplitude-dispersion-files",
        nargs=argparse.ZERO_OR_MORE,
        help="Optional: List the paths of existing amplitude dispersion files.",
        default=[],
    )
    inputs.add_argument(
        "--mask-file",
        help=(
            "Path to Byte mask file used to ignore low correlation/bad data (e.g water"
            " mask). Convention is 0 for no data/invalid, and 1 for good data."
        ),
    )

    # Phase linking options
    pl_group = parser.add_argument_group("Phase Linking options")
    pl_group.add_argument(
        "-ms",
        "--ministack-size",
        default=15,
        help="Strides/decimation factor (x, y) (in pixels) to use when determining",
    )
    # Half window size for the phase linking algorithm
    pl_group.add_argument(
        "-hw",
        "--half-window-size",
        type=int,
        nargs=2,
        default=(11, 5),
        metavar=("X", "Y"),
        help="Half window size for the phase linking algorithm",
    )
    pl_group.add_argument(
        "--shp-method",
        type=ShpMethod,
        choices=[s.value for s in ShpMethod],
        default=ShpMethod.GLRT,
        help="Method used to calculate the SHP.",
    )
    pl_group.add_argument(
        "--baseline-lag",
        type=Optional[int],
        default=None,
        help="Half window size for the phase linking algorithm",
    )

    # PS options
    ps_group = parser.add_argument_group("PS options")
    ps_group.add_argument(
        "--amp-dispersion-threshold",
        type=float,
        default=0.25,
        help="Threshold for the amplitude dispersion.",
    )

    # Unwrap options
    unwrap_group = parser.add_argument_group("Unwrap options")
    unwrap_group.add_argument(
        "--no-unwrap", action="store_true", help="Skip running the unwrapping step."
    )
    unwrap_group.add_argument(
        "--unwrap-method",
        choices=[m.value for m in UnwrapMethod],
        default=UnwrapMethod.SNAPHU,
        help="Choice of unwrapping algorithm to use.",
    )
    unwrap_group.add_argument(
        "-t",
        "--ntiles",
        nargs=2,
        type=int,
        metavar=("row_tiles", "col_tiles"),
        default=(1, 1),
        help=(
            "(For multiscale unwrapping) Number of tiles to split interferograms into."
        ),
    )
    unwrap_group.add_argument(
        "--downsample-factor",
        nargs=2,
        type=int,
        metavar=("row_downsample", "col_downsample"),
        default=(1, 1),
        help=(
            "(For multiscale unwrapping) Factor to multilook the coarse unwrapped"
            " version."
        ),
    )
    unwrap_group.add_argument(
        "--n-parallel-unwrap",
        type=int,
        default=1,
        help="Number of interferograms to unwrap in parallel.",
    )
    unwrap_group.add_argument(
        "--run-goldstein",
        action="store_true",
        help="Run Goldstein filter before unwrapping.",
    )
    unwrap_group.add_argument(
        "--run-interpolation",
        action="store_true",
        help="Run interpolation before unwrapping.",
    )

    # Correction options
    correction_group = parser.add_argument_group("Correction options")
    correction_group.add_argument(
        "--troposphere-files",
        nargs=argparse.ZERO_OR_MORE,
        help="List the paths of all troposphere files to include.",
    )
    correction_group.add_argument(
        "--tropo-date-fmt",
        default="%Y%m%d",
        help="Format of dates contained in weather-model filenames.",
    )
    correction_group.add_argument(
        "--tropo-package",
        default="pyaps",
        help="Package to use for tropospheric correction. Choices are: pyaps, raider.",
    )
    correction_group.add_argument(
        "--tropo-model",
        default=TropoModel.ERA5.value,
        choices=[t.value for t in TropoModel],
        type=TropoModel,
        help="source of the atmospheric model.",
    )
    correction_group.add_argument(
        "--tropo-delay_type",
        default=TropoType.COMB.value,
        type=TropoType,
        help="Tropospheric delay type: comb contains both wet and dry delays",
        choices=[t.value for t in TropoType],
    )
    correction_group.add_argument(
        "--ionosphere-files",
        nargs=argparse.ZERO_OR_MORE,
        help="List the paths of all ionosphere files to include.",
    )
    correction_group.add_argument(
        "--geometry-files",
        nargs=argparse.ZERO_OR_MORE,
        help="List the paths of all geometry files to include.",
    )
    correction_group.add_argument(
        "--dem-file",
        help="Path to DEM file to use for topographic correction.",
    )

    # Get Outputs from the command line
    out_group = parser.add_argument_group("Output options")
    out_group.add_argument(
        "--single-update",
        action="store_true",
        help="Create only one interferogram from the first and last SLC images.",
    )
    out_group.add_argument(
        "-s",
        "--strides",
        nargs=2,
        type=int,
        metavar=("X", "Y"),
        default=(1, 1),
        help=(
            "Strides/decimation factor (x, y) (in pixels) to use when determining"
            " output shape."
        ),
    )
    out_group.add_argument(
        "--output-bounds",
        nargs=4,
        type=float,
        metavar=("left", "bottom", "right", "top"),
        help="Requested bounding box (in lat/lon) for final output.",
    )

    worker_group = parser.add_argument_group("Worker options")
    worker_group.add_argument(
        "--log-file", help="Path to log to, in addition to stderr"
    )
    worker_group.add_argument(
        "--work-directory",
        type=Path,
        default=Path().resolve(),
        help="Path to directory to store intermediate/output files.",
    )
    worker_group.add_argument(
        "--enable-gpu",
        action="store_true",
        help="Enable usage of GPU (if using a machine that has one available).",
    )
    worker_group.add_argument(
        "--block-shape",
        type=int,
        nargs=2,
        default=(512, 512),
        help="Shape (rows, col) of blocks of data to load at once time.",
    )
    worker_group.add_argument(
        "--n-parallel-bursts",
        type=int,
        default=1,
        help="Number of bursts to process in parallel.",
    )
    worker_group.add_argument(
        "--threads-per-worker",
        type=int,
        default=max(1, get_cpu_count() // 2),
        help="Number of threads to use per worker.",
    )
    parser.set_defaults(run_func=create_config)

    return parser



"""Package for phase linking stacks of SLCs.

Currently implements the eigenvalue-based maximum likelihood (EMI) algorithm from
[@Ansari2018EfficientPhaseEstimation], as well as the EVD based approach from
[@Fornaro2015CAESARApproachBased] and [@Mirzaee2023NonlinearPhaseLinking]
"""

from ._compress import compress
from ._core import PhaseLinkRuntimeError, run_phase_linking



import argparse
from pathlib import Path
from typing import TYPE_CHECKING, Any

from dolphin.filtering import filter_rasters

if TYPE_CHECKING:
    _SubparserType = argparse._SubParsersAction[argparse.ArgumentParser]
else:
    _SubparserType = Any


def get_parser(subparser=None, subcommand_name="unwrap") -> argparse.ArgumentParser:
    """Set up the command line interface."""
    metadata = {
        "description": (
            "Filter unwrapped interferograms using a long-wavelength filter."
        ),
        "formatter_class": argparse.ArgumentDefaultsHelpFormatter,
        # https://docs.python.org/3/library/argparse.html#fromfile-prefix-chars
        "fromfile_prefix_chars": "@",
    }
    if subparser:
        # Used by the subparser to make a nested command line interface
        parser = subparser.add_parser(subcommand_name, **metadata)
    else:
        parser = argparse.ArgumentParser(**metadata)  # type: ignore[arg-type]

    # parser._action_groups.pop()
    parser.add_argument(
        "-o",
        "--output-dir",
        type=Path,
        help=(
            "Path to output directory to store results. None stores in same location as"
            " inputs"
        ),
    )
    # Get Inputs from the command line
    inputs = parser.add_argument_group("Input options")
    inputs.add_argument(
        "--unw-filenames",
        nargs=argparse.ONE_OR_MORE,
        type=Path,
        help=(
            "List the paths of unwrapped files to filter. Can pass a newline delimited"
            " file with @ifg_filelist.txt"
        ),
    )
    inputs.add_argument(
        "--temporal-coherence-filename",
        type=Path,
        help="Optionally, list the path of the temporal coherence to mask.",
    )
    inputs.add_argument(
        "--cor-filenames",
        nargs=argparse.ZERO_OR_MORE,
        help="Optionally, list the paths of the correlation files to use for masking",
    )
    inputs.add_argument(
        "--conncomp-filenames",
        nargs=argparse.ZERO_OR_MORE,
        help="Optionally, list the paths of the connected component labels for masking",
    )
    parser.add_argument(
        "--wavelength-cutoff",
        type=float,
        default=50_000,
        help="Spatial wavelength_cutoff (in meters) of filter to use.",
    )

    parser.add_argument(
        "--max-workers",
        type=int,
        default=1,
        help="Number of parallel files to filter.",
    )

    parser.set_defaults(run_func=_run_filter)

    return parser


def _run_filter(*args, **kwargs):
    """Run `dolphin.filtering.filter_long_wavelength`."""
    return filter_rasters(*args, **kwargs)


def main(args=None):
    """Get the command line arguments and filter files."""
    parser = get_parser()
    parsed_args = parser.parse_args(args)
    return filter_rasters(**vars(parsed_args))


if __name__ == "__main__":
    main()



import numpy as np
import pytest

from dolphin import similarity

# Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.
pytestmark = pytest.mark.filterwarnings(
    "ignore::rasterio.errors.NotGeoreferencedWarning",
)


def test_get_circle_idxs():
    idxs = similarity.get_circle_idxs(3)
    expected = np.array(
        [
            [-2, -1],
            [-2, 0],
            [-2, 1],
            [-1, -2],
            [-1, -1],
            [-1, 0],
            [-1, 1],
            [-1, 2],
            [0, -2],
            [0, -1],
            [0, 1],
            [0, 2],
            [1, -2],
            [1, -1],
            [1, 0],
            [1, 1],
            [1, 2],
            [2, -1],
            [2, 0],
            [2, 1],
        ]
    )
    np.testing.assert_array_equal(idxs, expected)


def test_pixel_similarity():
    x1, x2 = np.exp(1j * np.angle(np.random.randn(2, 50) + np.random.randn(2, 50) * 1j))
    sim = similarity.phase_similarity(x1[:5], x2[:5])
    assert -1 <= sim <= 1

    # For random noise, the similarity should be closer to zero with more data
    sim_full = similarity.phase_similarity(x1, x2)
    assert np.abs(sim_full) < np.abs(sim)

    # self similarity == 1
    assert similarity.phase_similarity(x1, x1) == 1


class TestMedianSimilarity:
    @pytest.fixture
    def ifg_stack(self, slc_stack):
        return slc_stack * slc_stack[[0]].conj()

    @pytest.mark.parametrize("radius", [2, 5, 9])
    def test_basic(self, ifg_stack, radius):
        sim = similarity.median_similarity(ifg_stack, search_radius=radius)
        assert np.all(sim > -1)
        assert np.all(sim < 1)

    @pytest.mark.parametrize("radius", [2, 5, 9])
    def test_median_similarity_masked(self, ifg_stack, radius):
        rows, cols = ifg_stack.shape[-2:]
        mask = np.random.rand(rows, cols).round().astype(bool)
        sim = similarity.median_similarity(ifg_stack, search_radius=radius, mask=mask)
        assert np.all(sim > -1)
        assert np.all(sim < 1)

    def test_create_similarity(self, tmp_path, slc_file_list):
        outfile = tmp_path / "med_sim.tif"
        similarity.create_similarities(
            slc_file_list, output_file=outfile, num_threads=1, block_shape=(64, 64)
        )


class TestMaxSimilarity:
    @pytest.fixture
    def ifg_stack(self, slc_stack):
        return slc_stack * slc_stack[[0]].conj()

    @pytest.mark.parametrize("radius", [2, 5, 9])
    def test_basic(self, ifg_stack, radius):
        sim = similarity.max_similarity(ifg_stack, search_radius=radius)
        assert np.all(sim > -1)
        assert np.all(sim < 1)

    @pytest.mark.parametrize("radius", [2, 5, 9])
    def test_max_similarity_masked(self, ifg_stack, radius):
        rows, cols = ifg_stack.shape[-2:]
        mask = np.random.rand(rows, cols).round().astype(bool)
        sim = similarity.max_similarity(ifg_stack, search_radius=radius, mask=mask)
        assert np.all(sim >= -1)
        assert np.all(sim <= 1)



import importlib.util
import os
from pathlib import Path

import numpy as np
import pytest

import dolphin.unwrap
from dolphin import io
from dolphin.workflows import SpurtOptions, TophuOptions, UnwrapMethod, UnwrapOptions

TOPHU_INSTALLED = importlib.util.find_spec("tophu") is not None
SPURT_INSTALLED = importlib.util.find_spec("spurt") is not None
WHIRLWIND_INSTALLED = importlib.util.find_spec("whirlwind") is not None


# Dataset has no geotransform, gcps, or rpcs. The identity matrix will be returned.
pytestmark = pytest.mark.filterwarnings(
    "ignore::rasterio.errors.NotGeoreferencedWarning",
    "ignore:.*io.FileIO.*:pytest.PytestUnraisableExceptionWarning",
    # TODO: Remove this when spurt removes `fork`:
    # RuntimeWarning: os.fork() was called. ...
    # ... unraisableexception.py:80: PytestUnraisableExceptionWarning
    "ignore::pytest.PytestUnraisableExceptionWarning",
)


@pytest.fixture()
def corr_raster(raster_100_by_200):
    # Make a correlation raster of all 1s in the same directory as the raster
    d = Path(raster_100_by_200).parent
    corr_raster = d / "corr_raster.cor.tif"
    array = np.ones((100, 200), dtype=np.float32)
    array[70, 100] = 0
    io.write_arr(
        arr=array,
        output_name=corr_raster,
        like_filename=raster_100_by_200,
        driver="GTiff",
    )
    return corr_raster


@pytest.fixture()
def unwrap_options():
    return UnwrapOptions()


class TestUnwrapSingle:
    def test_unwrap_snaphu_default(self, tmp_path, list_of_gtiff_ifgs, corr_raster):
        unw_filename = tmp_path / "unwrapped.unw.tif"
        unw_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=list_of_gtiff_ifgs[0],
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            nlooks=1,
        )
        assert unw_path == unw_filename
        assert str(conncomp_path) == str(unw_filename).replace(
            ".unw.tif", ".unw.conncomp.tif"
        )
        assert io.get_raster_xysize(unw_filename) == io.get_raster_xysize(
            list_of_gtiff_ifgs[0]
        )

    @pytest.mark.parametrize("init_method", ["mst", "mcf"])
    def test_unwrap_snaphu(
        self, tmp_path, list_of_gtiff_ifgs, corr_raster, init_method, unwrap_options
    ):
        # test other init_method
        unw_filename = tmp_path / "unwrapped.unw.tif"
        unwrap_options.snaphu_options.init_method = init_method
        unw_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=list_of_gtiff_ifgs[0],
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            nlooks=1,
            unwrap_options=unwrap_options,
        )
        assert unw_path.exists()
        assert conncomp_path.exists()
        assert io.get_raster_nodata(unw_path) == 0
        assert io.get_raster_nodata(conncomp_path) == 65535

    @pytest.mark.parametrize("method", [UnwrapMethod.ICU, UnwrapMethod.PHASS])
    def test_unwrap_methods(self, tmp_path, raster_100_by_200, corr_raster, method):
        unw_filename = tmp_path / f"{method.value}_unwrapped.unw.tif"
        unwrap_options = UnwrapOptions(unwrap_method=method)
        u_path, c_path = dolphin.unwrap.unwrap(
            ifg_filename=raster_100_by_200,
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            nlooks=1,
            unwrap_options=unwrap_options,
        )
        assert u_path.exists()
        assert c_path.exists()

    def test_unwrap_logfile(self, tmp_path, raster_100_by_200, corr_raster):
        unw_filename = tmp_path / "unwrapped.unw.tif"
        unwrap_options = UnwrapOptions(unwrap_method="icu")
        u_path, c_path = dolphin.unwrap.unwrap(
            ifg_filename=raster_100_by_200,
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            unwrap_options=unwrap_options,
            nlooks=1,
            log_to_file=True,
        )
        logfile_name = str(unw_filename).replace(".unw.tif", ".unw.log")
        assert Path(logfile_name).exists()
        assert u_path.exists()
        assert c_path.exists()

    def test_unwrap_snaphu_nodata(
        self, tmp_path, list_of_gtiff_ifgs, corr_raster, unwrap_options
    ):
        # test other init_method
        unw_filename = tmp_path / "unwrapped.unw.tif"
        unw_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=list_of_gtiff_ifgs[0],
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            unwrap_options=unwrap_options,
            nlooks=1,
            ccl_nodata=123,
            unw_nodata=np.nan,
        )
        assert unw_path.exists()
        assert conncomp_path.exists()
        assert np.isnan(io.get_raster_nodata(unw_path))
        assert io.get_raster_nodata(conncomp_path) == 123

    def test_goldstein(
        self,
        tmp_path,
        list_of_gtiff_ifgs,
        corr_raster,
        unwrap_options,
    ):
        # test other init_method
        unw_filename = tmp_path / "unwrapped.unw.tif"
        scratch_dir = tmp_path / "scratch"
        scratch_dir.mkdir()

        unwrap_options.run_goldstein = True
        unw_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=list_of_gtiff_ifgs[0],
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            nlooks=1,
            unwrap_options=unwrap_options,
            scratchdir=scratch_dir,
        )
        assert unw_path.exists()
        assert conncomp_path.exists()
        # Check there are no extraneous ".filter" files
        unw_dir = Path(unw_path).parent
        assert set(unw_dir.glob("*.unw.tif")) == {unw_path}

    @pytest.mark.parametrize("method", [UnwrapMethod.SNAPHU, UnwrapMethod.PHASS])
    @pytest.mark.parametrize("run_interpolation", [False, True])
    @pytest.mark.parametrize("run_goldstein", [False, True])
    def test_multiple_preprocess(
        self,
        tmp_path,
        list_of_gtiff_ifgs,
        corr_raster,
        unwrap_options,
        method,
        run_interpolation,
        run_goldstein,
    ):
        # test other init_method
        unw_filename = tmp_path / "unwrapped.unw.tif"
        scratch_dir = tmp_path / "scratch"
        scratch_dir.mkdir()
        unwrap_options.unwrap_method = method
        unwrap_options.run_interpolation = run_interpolation
        unwrap_options.run_goldstein = run_goldstein
        unw_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=list_of_gtiff_ifgs[0],
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            nlooks=1,
            unwrap_options=unwrap_options,
            scratchdir=scratch_dir,
        )
        assert unw_path.exists()
        assert conncomp_path.exists()
        # Check there are no extraneous ".interp" files
        unw_dir = Path(unw_path).parent
        assert set(unw_dir.glob("*.unw.tif")) == {unw_path}

    def test_interp_loop(self):
        x, y = np.meshgrid(np.arange(200), np.arange(100))
        # simulate a simple phase ramp
        phase = 0.003 * x + 0.002 * y
        # interferogram with the simulated phase ramp and with a constant amplitude
        ifg = np.exp(1j * phase)
        corr = np.ones(ifg.shape)
        # mask out the ifg/corr at a given pixel for example at pixel 50,40
        # index of the pixel of interest
        x_idx = 50
        y_idx = 40
        corr[y_idx, x_idx] = 0
        # generate indices for the pixels to be used in interpolation
        indices = np.array(
            dolphin.similarity.get_circle_idxs(
                max_radius=51, min_radius=0, sort_output=False
            )
        )
        # interpolate pixels with zero in the corr and write to interpolated_ifg
        interpolated_ifg = np.zeros((100, 200), dtype=np.complex64)
        dolphin.interpolation._interp_loop(
            ifg,
            corr,
            weight_cutoff=0.5,
            num_neighbors=20,
            alpha=0.75,
            indices=indices,
            interpolated_ifg=interpolated_ifg,
        )
        # expected phase based on the model above used for simulation
        expected_phase = 0.003 * x_idx + 0.002 * y_idx
        phase_error = np.angle(
            interpolated_ifg[y_idx, x_idx] * np.exp(-1j * expected_phase)
        )
        assert np.allclose(phase_error, 0.0, atol=1e-3)

    @pytest.mark.parametrize("method", [UnwrapMethod.SNAPHU, UnwrapMethod.PHASS])
    def test_interpolation(self, tmp_path, list_of_gtiff_ifgs, corr_raster, method):
        # test other init_method
        unw_filename = tmp_path / "unwrapped.unw.tif"
        scratch_dir = tmp_path / "scratch"
        scratch_dir.mkdir()
        unwrap_options = UnwrapOptions(unwrap_method=method, run_interpolation=True)
        unw_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=list_of_gtiff_ifgs[0],
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            nlooks=1,
            unwrap_options=unwrap_options,
            scratchdir=scratch_dir,
        )
        assert unw_path.exists()
        assert conncomp_path.exists()
        # Check there are no extraneous ".interp" files
        unw_dir = Path(unw_path).parent
        assert set(unw_dir.glob("*.unw.tif")) == {unw_path}


class TestUnwrapRun:
    def test_run_gtiff(self, list_of_gtiff_ifgs, corr_raster, unwrap_options):
        ifg_path = list_of_gtiff_ifgs[0].parent
        u_paths, c_paths = dolphin.unwrap.run(
            ifg_filenames=list_of_gtiff_ifgs,
            cor_filenames=[corr_raster] * len(list_of_gtiff_ifgs),
            output_path=ifg_path,
            nlooks=1,
            unwrap_options=unwrap_options,
        )
        assert all(p.exists() for p in u_paths)
        assert all(p.exists() for p in c_paths)

    def test_run_envi(self, list_of_envi_ifgs, corr_raster, unwrap_options):
        ifg_path = list_of_envi_ifgs[0].parent
        unwrap_options.snaphu_options.init_method = "mcf"
        u_paths, c_paths = dolphin.unwrap.run(
            ifg_filenames=list_of_envi_ifgs,
            cor_filenames=[corr_raster] * len(list_of_envi_ifgs),
            output_path=ifg_path,
            nlooks=1,
            unwrap_options=unwrap_options,
        )
        assert all(p.exists() for p in u_paths)
        assert all(p.exists() for p in c_paths)


class TestTophu:
    @pytest.mark.skipif(
        not TOPHU_INSTALLED, reason="tophu not installed for multiscale unwrapping"
    )
    def test_unwrap_multiscale(self, tmp_path, raster_100_by_200, corr_raster):
        unw_filename = tmp_path / "unwrapped.unw.tif"

        to = TophuOptions(ntiles=(2, 2), downsample_factor=(3, 3))
        unwrap_options = UnwrapOptions(unwrap_method="phass", tophu_options=to)
        out_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=raster_100_by_200,
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            unwrap_options=unwrap_options,
            nlooks=1,
        )
        assert out_path.exists()
        assert conncomp_path.exists()

    @pytest.mark.skipif(
        not TOPHU_INSTALLED, reason="tophu not installed for multiscale unwrapping"
    )
    def test_unwrap_multiscale_callback_given(
        self, tmp_path, raster_100_by_200, corr_raster
    ):
        from tophu import ICUUnwrap

        unw_filename = tmp_path / "unwrapped.unw.tif"
        unwrap_callback = ICUUnwrap()
        out_path, conncomp_path = dolphin.unwrap.multiscale_unwrap(
            ifg_filename=raster_100_by_200,
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            downsample_factor=(2, 2),
            ntiles=(2, 2),
            unwrap_callback=unwrap_callback,
            unw_nodata=0,
            nlooks=1,
        )
        assert out_path.exists()
        assert conncomp_path.exists()


class TestSpurt:
    @pytest.fixture()
    def ifg_file_list(self, tmp_path, slc_stack, slc_date_list):
        from dolphin import io
        from dolphin.phase_link import simulate

        slc_stack = np.exp(
            1j * simulate.make_defo_stack((20, 100, 200), sigma=1)
        ).astype("complex64")
        ifg_stack = slc_stack[1:] * slc_stack[[0]].conj()
        # Write to a file
        d = tmp_path / "gtiff"
        d.mkdir()
        name_template = d / f"{slc_date_list[0].strftime('%Y%m%d')}_{{date}}.slc.tif"

        file_list = []
        for cur_date, cur_ifg in zip(slc_date_list[1:], ifg_stack):
            fname = str(name_template).format(date=cur_date.strftime("%Y%m%d"))
            file_list.append(Path(fname))
            io.write_arr(arr=cur_ifg, output_name=fname)

        return file_list

    @pytest.mark.skipif(
        not SPURT_INSTALLED, reason="spurt not installed for 3d unwrapping"
    )
    def test_unwrap_spurt(self, tmp_path, ifg_file_list, corr_raster):
        opts = SpurtOptions()
        unwrap_options = UnwrapOptions(unwrap_method="spurt", spurt_options=opts)
        out_paths, conncomp_paths = dolphin.unwrap.run(
            ifg_filenames=ifg_file_list,
            cor_filenames=ifg_file_list,  # NOT USED... but required for `run`?
            temporal_coherence_file=corr_raster,
            unwrap_options=unwrap_options,
            output_path=tmp_path,
            nlooks=5,
        )
        assert all(p.exists() for p in out_paths)
        # spurt gives 0 conncomps for now:
        # TODO: Uncomment this if spurt starts making conncomps
        # assert all(p.exists() for p in conncomp_paths)


@pytest.mark.skipif(not WHIRLWIND_INSTALLED, reason="whirlwind package not installed")
class TestWhirlwind:
    def test_unwrap_whirlwind(self, tmp_path, raster_100_by_200, corr_raster):
        unw_filename = tmp_path / "whirlwind-unwrapped.unw.tif"

        unwrap_options = UnwrapOptions(unwrap_method="whirlwind")
        out_path, conncomp_path = dolphin.unwrap.unwrap(
            ifg_filename=raster_100_by_200,
            corr_filename=corr_raster,
            unw_filename=unw_filename,
            unwrap_options=unwrap_options,
            nlooks=1,
        )
        assert out_path.exists()
        assert conncomp_path.exists()


@pytest.mark.skipif(os.environ.get("NUMBA_DISABLE_JIT") == "1", reason="JIT disabled")
def test_compute_phase_diffs():
    # test on a 2D array with no phase jumps > pi
    phase1 = np.array([[0, 1], [1, 2]], dtype=float)
    expected1 = np.array([[0, 0], [0, 0]], dtype=float)
    assert np.allclose(dolphin.unwrap.compute_phase_diffs(phase1), expected1)

    # test on a 2D array with some phase jumps > pi at the top-left pixel
    phase2 = np.array([[0, 3.15], [3.15, 0]], dtype=float)
    expected2 = np.array([[2, 0], [0, 0]], dtype=float)
    assert np.allclose(dolphin.unwrap.compute_phase_diffs(phase2), expected2)

    # test on a larger 2D array
    phase3 = np.full((10, 10), np.pi, dtype=float)
    expected3 = np.zeros((10, 10), dtype=float)
    assert np.allclose(dolphin.unwrap.compute_phase_diffs(phase3), expected3)



from __future__ import annotations

import datetime
import logging
import time
from pathlib import Path
from typing import Optional, Sequence, cast

import numpy as np
from opera_utils import get_dates, make_nodata_mask

from dolphin import Bbox, Filename, interferogram, masking, ps, stack
from dolphin._log import log_runtime, setup_logging
from dolphin.io import VRTStack

from . import InterferogramNetwork, sequential
from .config import DisplacementWorkflow

logger = logging.getLogger(__name__)


@log_runtime
def run(
    cfg: DisplacementWorkflow, debug: bool = False, tqdm_kwargs=None
) -> tuple[list[Path], list[Path], Path, Path, Path, Path]:
    """Run the displacement workflow on a stack of SLCs.

    Parameters
    ----------
    cfg : DisplacementWorkflow
        [`DisplacementWorkflow`][dolphin.workflows.config.DisplacementWorkflow] object
        for controlling the workflow.
    debug : bool, optional
        Enable debug logging, by default False.
    tqdm_kwargs : dict, optional
        dict of arguments to pass to `tqdm` (e.g. `position=n` for n parallel bars)
        See https://tqdm.github.io/docs/tqdm/#tqdm-objects for all options.

    Returns
    -------
    ifg_file_list : list[Path]
        list of Paths to virtual interferograms created.
    comp_slc_file_list : list[Path]
        Paths to the compressed SLC files created from each ministack.
    temp_coh_file : Path
        Path to temporal coherence file created.
        In the case of a single phase linking step, this is from one phase linking step.
        In the case of sequential phase linking, this is the average of all ministacks.
    ps_looked_file : Path
        The multilooked boolean persistent scatterer file.
    amp_disp_looked_file : Path
        The multilooked amplitude dispersion file.
    shp_count_file : Path
        Path to the created SHP counts file.

    """
    t0 = time.perf_counter()
    setup_logging(debug=debug, filename=cfg.log_file)
    if tqdm_kwargs is None:
        tqdm_kwargs = {}
    work_dir = cfg.work_directory
    logger.info("Running wrapped phase estimation in %s", work_dir)

    input_file_list = cfg.cslc_file_list

    # #############################################
    # Make a VRT pointing to the input SLC files
    # #############################################
    subdataset = cfg.input_options.subdataset
    vrt_stack = VRTStack(
        input_file_list,
        subdataset=subdataset,
        outfile=cfg.work_directory / "slc_stack.vrt",
    )

    # Mark any files beginning with "compressed" as compressed
    is_compressed = ["compressed" in str(f).lower() for f in input_file_list]
    input_dates = _get_input_dates(
        input_file_list, is_compressed, cfg.input_options.cslc_date_fmt
    )
    reference_date, reference_idx = _get_reference_date_idx(
        input_file_list, is_compressed, input_dates
    )

    non_compressed_slcs = [
        f for f, is_comp in zip(input_file_list, is_compressed) if not is_comp
    ]

    # Create a mask file from input bounding polygons and/or specified output bounds
    mask_filename = _get_mask(
        output_dir=cfg.work_directory,
        output_bounds=cfg.output_options.bounds,
        like_filename=vrt_stack.outfile,
        cslc_file_list=non_compressed_slcs,
    )

    nodata_mask = masking.load_mask_as_numpy(mask_filename) if mask_filename else None
    # ###############
    # PS selection
    # ###############
    ps_output = cfg.ps_options._output_file
    ps_output.parent.mkdir(parents=True, exist_ok=True)
    if ps_output.exists():
        logger.info(f"Skipping making existing PS file {ps_output}")
    else:
        logger.info(f"Creating persistent scatterer file {ps_output}")
        try:
            existing_amp: Optional[Path] = cfg.amplitude_mean_files[0]
            existing_disp: Optional[Path] = cfg.amplitude_dispersion_files[0]
        except IndexError:
            existing_amp = existing_disp = None

        kwargs = tqdm_kwargs | {"desc": f"PS ({ps_output.parent})"}
        ps.create_ps(
            reader=vrt_stack,
            like_filename=vrt_stack.outfile,
            output_file=ps_output,
            output_amp_mean_file=cfg.ps_options._amp_mean_file,
            output_amp_dispersion_file=cfg.ps_options._amp_dispersion_file,
            amp_dispersion_threshold=cfg.ps_options.amp_dispersion_threshold,
            existing_amp_dispersion_file=existing_disp,
            nodata_mask=nodata_mask,
            existing_amp_mean_file=existing_amp,
            block_shape=cfg.worker_settings.block_shape,
            **kwargs,
        )

    # Save a looked version of the PS mask too
    strides = cfg.output_options.strides
    ps_looked_file, amp_disp_looked_file = ps.multilook_ps_files(
        strides=strides,
        ps_mask_file=cfg.ps_options._output_file,
        amp_dispersion_file=cfg.ps_options._amp_dispersion_file,
    )

    # #########################
    # phase linking/EVD step
    # #########################
    pl_path = cfg.phase_linking._directory
    pl_path.mkdir(parents=True, exist_ok=True)

    ministack_planner = stack.MiniStackPlanner(
        file_list=input_file_list,
        dates=input_dates,
        is_compressed=is_compressed,
        output_folder=pl_path,
        max_num_compressed=cfg.phase_linking.max_num_compressed,
        reference_date=reference_date,
        reference_idx=reference_idx,
    )

    phase_linked_slcs = sorted(pl_path.glob("2*.tif"))
    if len(phase_linked_slcs) > 0:
        logger.info(f"Skipping EVD step, {len(phase_linked_slcs)} files already exist")
        comp_slc_list = sorted(pl_path.glob("compressed*tif"))
        temp_coh_file = next(pl_path.glob("temporal_coherence*tif"))
        shp_count_file = next(pl_path.glob("shp_count*tif"))
    else:
        logger.info(f"Running sequential EMI step in {pl_path}")
        kwargs = tqdm_kwargs | {"desc": f"Phase linking ({pl_path})"}

        # TODO: Need a good way to store the nslc attribute in the PS file...
        # If we pre-compute it from some big stack, we need to use that for SHP
        # finding, not use the size of `slc_vrt_file`
        shp_nslc = None
        (phase_linked_slcs, comp_slc_list, temp_coh_file, shp_count_file) = (
            sequential.run_wrapped_phase_sequential(
                slc_vrt_file=vrt_stack.outfile,
                ministack_planner=ministack_planner,
                ministack_size=cfg.phase_linking.ministack_size,
                half_window=cfg.phase_linking.half_window.model_dump(),
                strides=strides,
                use_evd=cfg.phase_linking.use_evd,
                beta=cfg.phase_linking.beta,
                mask_file=mask_filename,
                ps_mask_file=ps_output,
                amp_mean_file=cfg.ps_options._amp_mean_file,
                amp_dispersion_file=cfg.ps_options._amp_dispersion_file,
                shp_method=cfg.phase_linking.shp_method,
                shp_alpha=cfg.phase_linking.shp_alpha,
                shp_nslc=shp_nslc,
                block_shape=cfg.worker_settings.block_shape,
                baseline_lag=cfg.phase_linking.baseline_lag,
                **kwargs,
            )
        )
    # Dump the used options for JSON parsing
    logger.info(
        "wrapped_phase complete",
        extra={
            "elapsed": time.perf_counter() - t0,
            "phase_linking_options": cfg.phase_linking.model_dump(mode="json"),
        },
    )

    # ###################################################
    # Form interferograms from estimated wrapped phase
    # ###################################################

    ifg_network = cfg.interferogram_network
    existing_ifgs = list(ifg_network._directory.glob("*.int.vrt"))
    if len(existing_ifgs) > 0:
        logger.info(f"Skipping interferogram step, {len(existing_ifgs)} exists")
        return (
            existing_ifgs,
            comp_slc_list,
            temp_coh_file,
            ps_looked_file,
            amp_disp_looked_file,
            shp_count_file,
        )

    logger.info(f"Creating virtual interferograms from {len(phase_linked_slcs)} files")
    ifg_file_list = create_ifgs(
        ifg_network, phase_linked_slcs, any(is_compressed), reference_date
    )
    return (
        ifg_file_list,
        comp_slc_list,
        temp_coh_file,
        ps_looked_file,
        amp_disp_looked_file,
        shp_count_file,
    )


def create_ifgs(
    interferogram_network: InterferogramNetwork,
    phase_linked_slcs: Sequence[Path],
    contained_compressed_slcs: bool,
    reference_date: datetime.datetime,
    dry_run: bool = False,
) -> list[Path]:
    """Create the list of interferograms for the `phase_linked_slcs`.

    Parameters
    ----------
    interferogram_network : InterferogramNetwork
        Parameters to determine which ifgs to form.
    phase_linked_slcs : Sequence[Path]
        Paths to phase linked SLCs.
    contained_compressed_slcs : bool
        Flag indicating that the inputs to phase linking contained compressed SLCs.
        Needed because the network must be handled differently if we started with
        compressed SLCs.
    reference_date : datetime.datetime
        Date/datetime of the "base phase" for the `phase_linked_slcs`
    dry_run : bool
        Flag indicating that the ifgs should not be written to disk.
        Default = False (ifgs will be created).

    Returns
    -------
    list[Path]
        List of output VRTInterferograms

    Raises
    ------
    ValueError
        If invalid parameters are passed which lead to 0 interferograms being formed
    NotImplementedError
        Currently raised for max-temporal-baseline networks when
        `contained_compressed_slcs` is True

    """
    ifg_dir = interferogram_network._directory
    if not dry_run:
        ifg_dir.mkdir(parents=True, exist_ok=True)
    ifg_file_list: list[Path] = []
    if not contained_compressed_slcs:
        # When no compressed SLCs were passed in to the config, we can directly pass
        # options to `Network` and get the ifg list
        network = interferogram.Network(
            slc_list=phase_linked_slcs,
            reference_idx=interferogram_network.reference_idx,
            max_bandwidth=interferogram_network.max_bandwidth,
            max_temporal_baseline=interferogram_network.max_temporal_baseline,
            indexes=interferogram_network.indexes,
            outdir=ifg_dir,
            write=not dry_run,
            verify_slcs=not dry_run,
        )
        if len(network.ifg_list) == 0:
            msg = "No interferograms were created"
            raise ValueError(msg)
        ifg_file_list = [ifg.path for ifg in network.ifg_list]  # type: ignore[misc]
        assert all(p is not None for p in ifg_file_list)

        return ifg_file_list

    # When we started with compressed SLCs, we need to do some extra work to get the
    # interferograms we want.
    # The total SLC phases we have to work with are
    # 1. reference date (might be before any dates in the filenames)
    # 2. the secondary of all phase-linked SLCs (which are the names of the files)

    # To get the ifgs from the reference date to secondary(conj), this involves doing
    # a `.conj()` on the phase-linked SLCs (which are currently `day1.conj() * day2`)
    single_ref_ifgs = [
        interferogram.convert_pl_to_ifg(
            f, reference_date=reference_date, output_dir=ifg_dir, dry_run=dry_run
        )
        for f in phase_linked_slcs
    ]

    # If we're only wanting single-reference day-(reference) to day-k interferograms,
    # these are all we need
    # XX Fix this hack for later
    if interferogram_network.indexes and interferogram_network.indexes == [(0, -1)]:
        ifg_file_list.append(single_ref_ifgs[-1])
        # # This isn't really what we want here, the logic is different than Network:
        # ifgs = [
        #     (single_ref_ifgs[ref_idx], single_ref_ifgs[sec_idx])
        #     for ref_idx, sec_idx in interferogram_network.indexes
        # ]
        # ifg_file_list.extend(ifgs)

    if interferogram_network.reference_idx == 0:
        ifg_file_list.extend(single_ref_ifgs)

    # For other networks, we have to combine other ones formed from the `Network`
    # Say we had inputs like:
    #  compressed_1_2_3 , slc_4, slc_5, slc_6
    # the compressed one is referenced to "1"
    # There will be 3 PL outputs for days 4, 5, 6, referenced to day "1":
    # (1, 4), (1, 5), (1, 6)
    # If we requested max-bw-2 interferograms, we want
    # (1, 4), (1, 5), (4, 5), (4, 6), (5, 6)
    # (the same as though we had normal SLCs (1, 4, 5, 6) )
    if interferogram_network.max_bandwidth is not None:
        max_b = interferogram_network.max_bandwidth
        # Max bandwidth is easier: take the first `max_b` from `phase_linked_slcs`
        # (which are the (ref_date, ...) interferograms),...
        ifgs_ref_date = single_ref_ifgs[:max_b]
        # ...then combine it with the results from the `Network`
        # Manually specify the dates, which come from the names of `phase_linked_slcs`
        secondary_dates = [get_dates(f)[0] for f in phase_linked_slcs]
        network_rest = interferogram.Network(
            slc_list=phase_linked_slcs,
            max_bandwidth=max_b,
            outdir=ifg_dir,
            dates=secondary_dates,
            write=not dry_run,
            verify_slcs=not dry_run,
        )
        # Using `cast` to assert that the paths are not None
        ifgs_others = cast(list[Path], [ifg.path for ifg in network_rest.ifg_list])

        ifg_file_list.extend(ifgs_ref_date + ifgs_others)

    if interferogram_network.max_temporal_baseline is not None:
        # Other types: TODO
        msg = (
            "max-temporal-baseline networks not yet supported when "
            " starting with compressed SLCs"
        )
        raise NotImplementedError(msg)

    # Dedupe, in case different options made the same ifg
    requested_ifgs = set(ifg_file_list)
    # remove ones we aren't using (in the case of a single index)
    written_ifgs = set(ifg_dir.glob("*.int*"))
    for p in written_ifgs - requested_ifgs:
        p.unlink()
    return ifg_file_list


def _get_reference_date_idx(
    input_file_list: Sequence[Path],
    is_compressed: Sequence[bool],
    input_dates: Sequence[Sequence[datetime.datetime]],
) -> tuple[datetime.datetime, int]:
    is_compressed = ["compressed" in str(f).lower() for f in input_file_list]
    if not is_compressed[0]:
        return input_dates[0][0], 0

    # Otherwise use the last Compressed SLC as reference
    reference_idx = np.where(is_compressed)[0][-1]
    reference_date = input_dates[reference_idx][0]

    return reference_date, reference_idx


def _get_input_dates(
    input_file_list: Sequence[Path], is_compressed: Sequence[bool], date_fmt: str
) -> list[list[datetime.datetime]]:
    input_dates = [get_dates(f, fmt=date_fmt) for f in input_file_list]
    # For any that aren't compressed, take the first date.
    # this is because the official product name of OPERA/Sentinel1 has both
    # "acquisition_date" ... "generation_date" in the filename
    # TODO: this is a bit hacky, perhaps we can make this some input option
    # so that the user can specify how to get dates from their files (or even
    # directly pass in dates?)
    return [
        dates[:1] if not is_comp else dates
        for dates, is_comp in zip(input_dates, is_compressed)
    ]


def _get_mask(
    output_dir: Path,
    output_bounds: Bbox | tuple[float, float, float, float] | None,
    like_filename: Filename,
    cslc_file_list: Sequence[Filename],
) -> Path | None:
    # Make the nodata mask from the polygons, if we're using OPERA CSLCs

    try:
        nodata_mask_file = output_dir / "nodata_mask.tif"
        make_nodata_mask(
            opera_file_list=cslc_file_list,
            out_file=nodata_mask_file,
            buffer_pixels=200,
        )
    except Exception as e:
        logger.warning(f"Could not make nodata mask: {e}")
        nodata_mask_file = None

    mask_filename: Path | None = None
    # Also mask outside the area of interest if we've specified a small bounds
    if output_bounds is not None:
        # Make a mask just from the bounds
        bounds_mask_filename = output_dir / "bounds_mask.tif"
        masking.create_bounds_mask(
            bounds=output_bounds,
            output_filename=bounds_mask_filename,
            like_filename=like_filename,
        )

        # Then combine with the nodata mask
        if nodata_mask_file is not None:
            combined_mask_filename = output_dir / "combined_mask.tif"
            masking.combine_mask_files(
                mask_files=[bounds_mask_filename, nodata_mask_file],
                output_file=combined_mask_filename,
                output_convention=masking.MaskConvention.ZERO_IS_NODATA,
            )
            mask_filename = combined_mask_filename
        else:
            mask_filename = bounds_mask_filename
    else:
        mask_filename = nodata_mask_file

    return mask_filename



from __future__ import annotations

# from dolphin._dates import get_dates, sort_files_by_date
import logging
from pathlib import Path
from typing import Any

from pydantic import ConfigDict, Field, field_validator

from ._common import (
    InputOptions,
    OutputOptions,
    WorkflowBase,
    _read_file_list_or_glob,
)
from ._unwrap_options import UnwrapOptions

__all__ = [
    "UnwrappingWorkflow",
]

logger = logging.getLogger(__name__)


class UnwrappingWorkflow(WorkflowBase):
    """Configuration for the unwrapping stage of the workflow."""

    # Paths to input/output files
    ifg_file_list: list[Path] = Field(
        default_factory=list,
        description=(
            "list of CSLC files, or newline-delimited file "
            "containing list of CSLC files."
        ),
    )
    input_options: InputOptions = Field(default_factory=InputOptions)
    output_options: OutputOptions = Field(default_factory=OutputOptions)

    unwrap_options: UnwrapOptions = Field(default_factory=UnwrapOptions)

    # internal helpers
    # Stores the list of directories to be created by the workflow
    model_config = ConfigDict(
        extra="allow", json_schema_extra={"required": ["ifg_file_list"]}
    )

    # validators
    # reuse the _read_file_list_or_glob
    _check_cslc_file_glob = field_validator("ifg_file_list", mode="before")(
        _read_file_list_or_glob
    )

    def model_post_init(self, __context: Any) -> None:
        """After validation, set up properties for use during workflow run."""
        super().model_post_init(__context)
        # Ensure outputs from workflow steps are within work directory.
        if not self.keep_paths_relative:
            # Resolve all CSLC paths:
            self.ifg_file_list = [p.resolve(strict=False) for p in self.ifg_file_list]

        work_dir = self.work_directory
        # move output dir inside the work directory (if it's not already inside).
        # They may already be inside if we're loading from a json/yaml file.
        opts = self.unwrap_options
        if opts._directory.parent != work_dir:
            opts._directory = work_dir / opts._directory
        if not self.keep_paths_relative:
            opts._directory = opts._directory.resolve(strict=False)

        # Track the directories that need to be created at start of workflow
        self._directory_list = [
            work_dir,
            self.unwrap_options._directory,
        ]



from dolphin import stack
from dolphin.io import _readers
from dolphin.phase_link import simulate
from dolphin.utils import gpu_is_available
from dolphin.workflows import single

GPU_AVAILABLE = gpu_is_available()
simulate._seed(1234)


def test_sequential_gtiff(tmp_path, slc_file_list):
    """Run through the sequential estimation with a GeoTIFF stack."""
    vrt_file = tmp_path / "slc_stack.vrt"
    files = slc_file_list[:3]
    vrt_stack = _readers.VRTStack(files, outfile=vrt_file)
    _, rows, cols = vrt_stack.shape
    is_compressed = [False] * len(files)
    ministack = stack.MiniStackInfo(
        file_list=vrt_stack.file_list,
        dates=vrt_stack.dates,
        is_compressed=is_compressed,
    )

    hy, hx = 1, 2
    half_window = {"x": hx, "y": hy}
    strides = {"x": 1, "y": 1}
    output_folder = tmp_path / "single"
    single.run_wrapped_phase_single(
        slc_vrt_file=vrt_file,
        ministack=ministack,
        output_folder=output_folder,
        half_window=half_window,
        strides=strides,
        shp_method="rect",
    )

    assert output_folder.exists()
    # Check that all the expected outputs are there
    assert len(list(output_folder.glob("2*.slc.tif"))) == 3
    assert len(list(output_folder.glob("compressed_*tif"))) == 1
    assert len(list(output_folder.glob("temporal_coherence*tif"))) == 1



"""Find the persistent scatterers in a stack of SLCS."""

from __future__ import annotations

import logging
import shutil
import warnings
from pathlib import Path
from typing import Optional, Sequence

import numpy as np
from numpy.typing import ArrayLike
from osgeo import gdal

from dolphin import io, utils
from dolphin._types import Filename
from dolphin.io import EagerLoader, StackReader, repack_raster

gdal.UseExceptions()

logger = logging.getLogger(__name__)

NODATA_VALUES = {"ps": 255, "amp_dispersion": 0.0, "amp_mean": 0.0}

FILE_DTYPES = {"ps": np.uint8, "amp_dispersion": np.float32, "amp_mean": np.float32}
_EXTRA_COMPRESSION = {
    "keep_bits": 10,
    "predictor": 3,
}
REPACK_OPTIONS = {
    "ps": {},
    "amp_dispersion": _EXTRA_COMPRESSION,
    "amp_mean": _EXTRA_COMPRESSION,
}


def create_ps(
    *,
    reader: StackReader,
    output_file: Filename,
    output_amp_mean_file: Filename,
    output_amp_dispersion_file: Filename,
    like_filename: Filename,
    amp_dispersion_threshold: float = 0.25,
    existing_amp_mean_file: Optional[Filename] = None,
    existing_amp_dispersion_file: Optional[Filename] = None,
    nodata_mask: Optional[np.ndarray] = None,
    update_existing: bool = False,
    block_shape: tuple[int, int] = (512, 512),
    **tqdm_kwargs,
):
    """Create the amplitude dispersion, mean, and PS files.

    Parameters
    ----------
    reader : StackReader
        A dataset reader for the 3D SLC stack.
    output_file : Filename
        The output PS file (dtype: Byte)
    output_amp_dispersion_file : Filename
        The output amplitude dispersion file.
    output_amp_mean_file : Filename
        The output mean amplitude file.
    like_filename : Filename
        The filename to use for the output files' spatial reference.
    amp_dispersion_threshold : float, optional
        The threshold for the amplitude dispersion. Default is 0.25.
    existing_amp_mean_file : Optional[Filename], optional
        An existing amplitude mean file to use, by default None.
    existing_amp_dispersion_file : Optional[Filename], optional
        An existing amplitude dispersion file to use, by default None.
    nodata_mask : Optional[np.ndarray]
        If provided, skips computing PS over areas where the mask is False
        Otherwise, loads input data from everywhere and calculates.
    update_existing : bool, optional
        If providing existing amp mean/dispersion files, combine them with the
        data from the current SLC stack.
        If False, simply uses the existing files to create as PS mask.
        Default is False.
    block_shape : tuple[int, int], optional
        The 2D block size to load all bands at a time.
        Default is (512, 512)
    **tqdm_kwargs : optional
        Arguments to pass to `tqdm`, (e.g. `position=n` for n parallel bars)
        See https://tqdm.github.io/docs/tqdm/#tqdm-objects for all options.

    """
    if existing_amp_dispersion_file and existing_amp_mean_file and not update_existing:
        logger.info("Using existing amplitude dispersion file, skipping calculation.")
        # Just use what's there, copy to the expected output locations
        _use_existing_files(
            existing_amp_mean_file=existing_amp_mean_file,
            existing_amp_dispersion_file=existing_amp_dispersion_file,
            output_file=output_file,
            output_amp_mean_file=output_amp_mean_file,
            output_amp_dispersion_file=output_amp_dispersion_file,
            amp_dispersion_threshold=amp_dispersion_threshold,
        )
        return

    # Otherwise, we need to calculate the PS files from the SLC stack
    # Initialize the output files with zeros
    file_list = [output_file, output_amp_dispersion_file, output_amp_mean_file]
    for fn, dtype, nodata in zip(
        file_list, FILE_DTYPES.values(), NODATA_VALUES.values()
    ):
        io.write_arr(
            arr=None,
            like_filename=like_filename,
            output_name=fn,
            nbands=1,
            dtype=dtype,
            nodata=nodata,
        )
    # Initialize the intermediate arrays for the calculation
    magnitude = np.zeros((reader.shape[0], *block_shape), dtype=np.float32)

    writer = io.BackgroundBlockWriter()
    # Make the generator for the blocks
    block_gen = EagerLoader(reader, block_shape=block_shape, nodata_mask=nodata_mask)
    for cur_data, (rows, cols) in block_gen.iter_blocks(**tqdm_kwargs):
        cur_rows, cur_cols = cur_data.shape[-2:]

        if not (np.all(cur_data == 0) or np.all(np.isnan(cur_data))):
            magnitude_cur = np.abs(cur_data, out=magnitude[:, :cur_rows, :cur_cols])
            mean, amp_disp, ps = calc_ps_block(
                # use min_count == size of stack so that ALL need to be not Nan
                magnitude_cur,
                amp_dispersion_threshold,
                min_count=len(magnitude_cur),
            )

            # Use the UInt8 type for the PS to save.
            # For invalid pixels, set to max Byte value
            ps = ps.astype(FILE_DTYPES["ps"])
            ps[amp_disp == 0] = NODATA_VALUES["ps"]
        else:
            # Fill the block with nodata
            ps = (
                np.ones((cur_rows, cur_cols), dtype=FILE_DTYPES["ps"])
                * NODATA_VALUES["ps"]
            )
            mean = np.full(
                (cur_rows, cur_cols),
                NODATA_VALUES["amp_mean"],
                dtype=FILE_DTYPES["amp_mean"],
            )
            amp_disp = np.full(
                (cur_rows, cur_cols),
                NODATA_VALUES["amp_dispersion"],
                dtype=FILE_DTYPES["amp_dispersion"],
            )

        # Write amp dispersion and the mean blocks
        writer.queue_write(mean, output_amp_mean_file, rows.start, cols.start)
        writer.queue_write(amp_disp, output_amp_dispersion_file, rows.start, cols.start)
        writer.queue_write(ps, output_file, rows.start, cols.start)

    logger.info(f"Waiting to write {writer.num_queued} blocks of data.")
    writer.notify_finished()
    # Repack for better compression
    logger.info("Repacking PS rasters for better compression")
    for fn, opt in zip(file_list, REPACK_OPTIONS.values()):
        # Repack to a temp, then overwrite
        repack_raster(Path(fn), output_dir=None, **opt)
    logger.info("Finished writing out PS files")


def calc_ps_block(
    stack_mag: ArrayLike,
    amp_dispersion_threshold: float = 0.25,
    min_count: Optional[int] = None,
):
    r"""Calculate the amplitude dispersion for a block of data.

    The amplitude dispersion is defined as the standard deviation of a pixel's
    magnitude divided by the mean of the magnitude:

    \[
    d_a = \frac{\sigma(|Z|)}{\mu(|Z|)}
    \]

    where $Z \in \mathbb{R}^{N}$ is one pixel's complex data for $N$ SLCs.

    Parameters
    ----------
    stack_mag : ArrayLike
        The magnitude of the stack of SLCs.
    amp_dispersion_threshold : float, optional
        The threshold for the amplitude dispersion to label a pixel as a PS:
            ps = amp_disp < amp_dispersion_threshold
        Default is 0.25.
    min_count : int, optional
        The minimum number of valid pixels to calculate the mean and standard
        deviation. If the number of valid pixels is less than `min_count`,
        then the mean and standard deviation are set to 0 (and the pixel is
        not a PS). Default is 90% the number of SLCs: `int(0.9 * stack_mag.shape[0])`.

    Returns
    -------
    mean : np.ndarray
        The mean amplitude for the block.
        dtype: float32
    amp_disp : np.ndarray
        The amplitude dispersion for the block.
        dtype: float32
    ps : np.ndarray
        The persistent scatterers for the block.
        dtype: bool

    Notes
    -----
    The min_count is used to prevent the mean and standard deviation from being
    calculated for pixels that are not valid for most of the SLCs. This happens
    when the burst footprints shift around and pixels near the edge get only one or
    two acquisitions.
    Since fewer samples are used to calculate the mean and standard deviation,
    there is a higher false positive risk for these edge pixels.

    """
    if np.iscomplexobj(stack_mag):
        msg = "The input `stack_mag` must be real-valued."
        raise ValueError(msg)

    if min_count is None:
        min_count = int(0.9 * stack_mag.shape[0])

    with warnings.catch_warnings():
        # ignore the warning about nansum/nanmean of empty slice
        warnings.simplefilter("ignore", category=RuntimeWarning)

        mean = np.nanmean(stack_mag, axis=0)
        std_dev = np.nanstd(stack_mag, axis=0)
        count = np.count_nonzero(~np.isnan(stack_mag), axis=0)
        amp_disp = std_dev / mean
    # Mask out the pixels with too few valid pixels
    amp_disp[count < min_count] = np.nan
    # replace nans/infinities with 0s, which will mean nodata
    mean = np.nan_to_num(mean, nan=0, posinf=0, neginf=0, copy=False)
    amp_disp = np.nan_to_num(amp_disp, nan=0, posinf=0, neginf=0, copy=False)

    ps = amp_disp < amp_dispersion_threshold
    ps[amp_disp == 0] = False
    return mean, amp_disp, ps


def _use_existing_files(
    *,
    existing_amp_mean_file: Filename,
    existing_amp_dispersion_file: Filename,
    output_file: Filename,
    output_amp_mean_file: Filename,
    output_amp_dispersion_file: Filename,
    amp_dispersion_threshold: float,
) -> None:
    amp_disp = io.load_gdal(existing_amp_dispersion_file, masked=True)
    ps = amp_disp < amp_dispersion_threshold
    ps = ps.astype(np.uint8)
    # Set the PS nodata value to the max uint8 value
    ps[(amp_disp == 0) | amp_disp.mask] = NODATA_VALUES["ps"]
    io.write_arr(
        arr=ps,
        like_filename=existing_amp_dispersion_file,
        output_name=output_file,
        nodata=NODATA_VALUES["ps"],
    )
    # Copy the existing amp mean file/amp dispersion file
    shutil.copy(existing_amp_dispersion_file, output_amp_dispersion_file)
    shutil.copy(existing_amp_mean_file, output_amp_mean_file)


def multilook_ps_files(
    strides: dict[str, int],
    ps_mask_file: Filename,
    amp_dispersion_file: Filename,
) -> tuple[Path, Path]:
    """Create a multilooked version of the full-res PS mask/amplitude dispersion.

    Parameters
    ----------
    strides : dict[str, int]
        Decimation factor for 'x', 'y'
    ps_mask_file : Filename
        Name of input full-res uint8 PS mask file
    amp_dispersion_file : Filename
        Name of input full-res float32 amplitude dispersion file

    Returns
    -------
    output_ps_file : Path
        Multilooked PS mask file
        Will be same as `ps_mask_file`, but with "_looked" added before suffix.
    output_amp_disp_file : Path
        Multilooked amplitude dispersion file
        Similar naming scheme to `output_ps_file`

    """
    if strides == {"x": 1, "y": 1}:
        logger.info("No striding request, skipping multilook.")
        return Path(ps_mask_file), Path(amp_dispersion_file)
    full_cols, full_rows = io.get_raster_xysize(ps_mask_file)
    out_rows, out_cols = full_rows // strides["y"], full_cols // strides["x"]

    ps_suffix = Path(ps_mask_file).suffix
    ps_out_path = Path(str(ps_mask_file).replace(ps_suffix, f"_looked{ps_suffix}"))
    logger.info(f"Saving a looked PS mask to {ps_out_path}")

    if Path(ps_out_path).exists():
        logger.info(f"{ps_out_path} exists, skipping.")
    else:
        ps_mask = io.load_gdal(ps_mask_file, masked=True).astype(bool)
        ps_mask_looked = utils.take_looks(
            ps_mask, strides["y"], strides["x"], func_type="any", edge_strategy="pad"
        )
        # make sure it's the same size as the MLE result/temp_coh after padding
        ps_mask_looked = ps_mask_looked[:out_rows, :out_cols]
        ps_mask_looked = ps_mask_looked.astype("uint8").filled(NODATA_VALUES["ps"])
        io.write_arr(
            arr=ps_mask_looked,
            like_filename=ps_mask_file,
            output_name=ps_out_path,
            strides=strides,
            nodata=NODATA_VALUES["ps"],
        )

    amp_disp_suffix = Path(amp_dispersion_file).suffix
    amp_disp_out_path = Path(
        str(amp_dispersion_file).replace(amp_disp_suffix, f"_looked{amp_disp_suffix}")
    )
    if amp_disp_out_path.exists():
        logger.info(f"{amp_disp_out_path} exists, skipping.")
    else:
        amp_disp = io.load_gdal(amp_dispersion_file, masked=True)
        # We use `nanmin` assuming that the multilooked PS is using
        # the strongest PS (the one with the lowest amplitude dispersion)
        amp_disp_looked = utils.take_looks(
            amp_disp,
            strides["y"],
            strides["x"],
            func_type="nanmin",
            edge_strategy="pad",
        )
        amp_disp_looked = amp_disp_looked[:out_rows, :out_cols]
        amp_disp_looked = amp_disp_looked.filled(NODATA_VALUES["amp_dispersion"])
        io.write_arr(
            arr=amp_disp_looked,
            like_filename=amp_dispersion_file,
            output_name=amp_disp_out_path,
            strides=strides,
            nodata=NODATA_VALUES["amp_dispersion"],
        )
    return ps_out_path, amp_disp_out_path


def combine_means(means: ArrayLike, N: ArrayLike) -> np.ndarray:
    r"""Compute the combined mean from multiple `mu_i` values.

    This function calculates the weighted average of amplitudes based on the
    number of original data points (N) that went into each mean.

    Parameters
    ----------
    means : ArrayLike
        A 3D array of mean values.
        Shape: (n_images, rows, cols)
    N : np.ndarray
        A list/array of weights indicating the number of original images.
        Shape: (depth,)

    Returns
    -------
    np.ndarray
        The combined mean.
        Shape: (height, width)

    Notes
    -----
    Both input arrays are expected to have the same shape.
    The operation is performed along axis=0.

    The combined mean is calculated as

    \begin{equation}
        E[X] = \frac{\sum_i N_i\mu_i}{\sum_i N_i}
    \end{equation}

    """
    N = np.asarray(N)
    if N.shape[0] != means.shape[0]:
        raise ValueError("Size of N must match the number of images in means.")
    if N.ndim == 1:
        N = N[:, None, None]

    weighted_sum = np.sum(means * N, axis=0)
    total_N = np.sum(N, axis=0)

    return weighted_sum / total_N


def combine_amplitude_dispersions(
    dispersions: np.ndarray, means: np.ndarray, N: ArrayLike | Sequence
) -> np.ndarray:
    r"""Compute the combined amplitude dispersion from multiple groups.

    Given several ADs where difference numbers of images, N, went in,
    the function computes a weighted mean/variance to calculate the combined AD.

    Parameters
    ----------
    dispersions : np.ndarray
        A 3D array of amplitude dispersion values for each group.
        Shape: (depth, height, width)
    means : np.ndarray
        A 3D array of mean values for each group.
        Shape: (depth, height, width)
    N : np.ndarray
        An array sample sizes for each group.
        Shape: (depth, )

    Returns
    -------
    np.ndarray
        The combined amplitude dispersion.
        Shape: (height, width)
    np.ndarray
        The combined amplitude mean.
        Shape: (height, width)

    Notes
    -----
    All input arrays are expected to have the same shape.
    The operation is performed along `axis=0`.

    Let $X_i$ be the random variable for group $i$, with mean $\mu_i$ and variance
    $\sigma_i^2$, and $N_i$ be the number of samples in group $i$.

    The combined variance $\sigma^2$ uses the formula

    \begin{equation}
        \sigma^2 = E[X^2] - (E[X])^2
    \end{equation}

    where $E[X]$ is the combined mean, and $E[X^2]$ is the expected value of
    the squared random variable.

    The combined mean is calculated as:

    \begin{equation}
        E[X] = \frac{\sum_i N_i\mu_i}{\sum_i N_i}
    \end{equation}

    For $E[X^2]$, we use the property $E[X^2] = \sigma^2 + \mu^2$:

    \begin{equation}
        E[X^2] = \frac{\sum_i N_i(\sigma_i^2 + \mu_i^2)}{\sum_i N_i}
    \end{equation}

    Substituting these into the variance formula gives:

    \begin{equation}
        \sigma^2 = \frac{\sum_i N_i(\sigma_i^2 + \mu_i^2)}{\sum_i N_i} -
        \left(\frac{\sum_i N_i\mu_i}{\sum_i N_i}\right)^2
    \end{equation}

    """
    N = np.asarray(N)
    if N.ndim == 1:
        N = N[:, None, None]
    if not (means.shape == dispersions.shape):
        raise ValueError("Input arrays must have the same shape.")
    if means.shape[0] != N.shape[0]:
        raise ValueError("Size of N must match the number of groups in means.")

    combined_mean = combine_means(means, N)

    # Compute combined variance
    variances = (dispersions * means) ** 2
    total_N = np.sum(N, axis=0).squeeze()
    sum_N_var_meansq = np.sum(N * (variances + means**2), axis=0)
    combined_variance = (sum_N_var_meansq / total_N) - combined_mean**2

    return np.sqrt(combined_variance) / combined_mean, combined_mean
